/**
  ******************************************************************************
  * @file   : Firmware.cs 
  * @author : Softweb Solutions An Avnet Company
  * @modify : 08-APR-2021
  * @brief  : Firmware part for DOTNET CORE SDK v3.0
  ******************************************************************************
*/

/**
 * Hope you have installed the DOTNET CORE SDK v3.0 as guided in README.md file or from documentation portal. 
 * Import the IoTConnect SDK package and other required packages
 */

using IOTConnectSDK;
using System;
using System.Threading.Tasks;
using Newtonsoft.Json;
using IOTConnectSDK.Models.Options;
using Newtonsoft.Json.Linq;
using System.Collections.Generic;
using Microsoft.Azure.Devices.Shared;

namespace IoTConnectSDK_Sample
{
    class Program
    {
        static SDKClient sdkClient;
        static void Main(string[] args)
        {
            /*
            ## Prerequisite parameter to run this sample code
            - cpId              :: It need to get from the IoTConnect platform "Settings->Key Vault".
            - uniqueId          :: Its device ID which register on IotConnect platform and also its status has Active and Acquired.
            - env               :: It need to get from the IoTConnect platform "Settings->Key Vault".
            - sdkOptions        :: It helps to define the path of self signed and CA signed certificate as well as define the offline storage params
            */
            string cpId = "<cpid>";
            string uniqueId = "<deviceUniqueId>";
            string env = "<env>";

            /** 
			* sdkOptions is optional. Mandatory for "certificate" X.509 device authentication type
			"certificate": It is indicated to define the path of the certificate file. Mandatory for X.509/SSL device CA signed or self-signed authentication type only.
				- SSL.Certificate: your device certificate
				- SSL.Password : certificate password
			"offlineStorage" : Define the configuration related to the offline data storage 
				- Disabled : false = offline data storing, true = not storing offline data 
				- AvailSpaceInMb : Define the file size of offline data which should be in (MB)
				- FileCount : Number of files need to create for offline data
			  Note: sdkOptions is optional but mandatory for SSL/x509 device authentication type only. Define proper setting or leave it NULL. If you do not provide offline storage, it will set the default settings as per defined above. It may harm your device by storing the large data. Once memory gets full may chance to stop the execution.
			*/

            SDKClientOption sdkOptions = new SDKClientOption();
            //For SSL CA signed and SelfSigned authorized device only
            sdkOptions.SSL.Certificate = "<<filepath>>/cert.pfx";
            sdkOptions.SSL.Password = "******";
            //For Offline Storage only
            sdkOptions.OfflineStorage.AvailSpaceInMb = << Integer Value >>; //size in MB, Default value = unlimited
            sdkOptions.OfflineStorage.FileCount = << Integer Value >>; // Default value = 1
            sdkOptions.OfflineStorage.Disabled = << true / false >>; // //default value = false, false = store data, true

            //Initialize device sdk client to connect device. 
            sdkClient = new SDKClient(cpId, uniqueId, DeviceCallback, TwinUpdateCallBack, sdkOptions, env);

            // Non Gateway
            string deviceData = "[{'uniqueId': '<< Device UniqueId >>','time' : '<< date >>','data': { }}]";

            // For Gateway and multiple child device 
            // string deviceData = "[{
            // 		'uniqueId': '<< Gateway Device UniqueId >>', // It should be must first object of the array
            // 		'time': '<< date >>',
            // 		'data': {}
            // 	},
            // 	{
            // 		'uniqueId':'<< Child DeviceId >>', //Child device
            // 		'time': '<< date >>',
            // 		'data': {}
            // 	}]";

            /**
			* Add your device attributes and respective value here as per standard format defined in sdk documentation
			* "time" : Date format should be as defined //"2021-01-24T10:06:17.857Z" 
			* "data" : JSON data type format // {"temperature": 15.55, "gyroscope" : { 'x' : -1.2 }}
			*/

            //Send data to device using SendData method.
            sdkClient.SendData(deviceData);

            /*
			Type    : Public Method "UpdateTwin()"
			Usage   : Update the twin reported property
			Input   : "key" and "value" as below
			Output  : 
			// string key = "<< Desired property key >>"; // Desired property key received from Twin callback message
			// string value = "<< Desired Property value >>"; // Value of respective desired property
			// sdkClient.UpdateTwin(key,value)

            /*
			Type    : Public Method "GetAllTwins()"
			Usage   : To get all the twin properties Desired and Reported
			Output  : All twin property will receive in callback function "TwinUpdateCallBack()"
			*/
            //	sdkClient.GetAllTwins();


            /*
			Type    : Public Method "GetAttributes()"
			Usage   : Send request to get all the attributes
			Input   :
			Output  :
			*/
            //  sdkClient.GetAttributes();

            /*
			Type    : Public Method "Dispose()"
			Usage   : Disconnect the device from cloud
			Note : It will disconnect the device after defined time
			//  sdkClient.Dispose();
			*/

            Console.ReadLine();
        }

        private static Task TwinUpdateCallBack(string arg)
        {
            try
            {
                Console.WriteLine("Twin Update received.");
                Console.WriteLine("Twin : " + arg);
                var twinDetails = JsonConvert.DeserializeObject<DeviceTwinCallBackModel>(arg);

                JObject twin = JObject.Parse(twinDetails.Desired.ToJson(Formatting.None));

                if (twin.Count > 0)
                {
                    Dictionary<string, object> twinDict = new Dictionary<string, object>();
                    foreach (JProperty item in twin.Children<JProperty>())
                    {
                        if (item.Name.Equals("$version", StringComparison.CurrentCultureIgnoreCase)) continue;
                        sdkClient.UpdateTwin(item.Name, Convert.ToString(item.Value));
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
            return Task.CompletedTask;
        }

        private static Task DeviceCallback(string message)
        {

            //"mt" = 5 - DeviceCommandAck, 11 - OtaAck/FirmwareUpgradeAck
            //"st" = 4 - Failed, 6 - ExecutedAck[DeviceCommandAck], 7 - Success[OtaAck/FirmwareUpgradeAck]
            try
            {
                var commandData = JsonConvert.DeserializeObject<CommandData>(message);

                DeviceAckModel ackDetails = new DeviceAckModel()
                {
                    AckId = commandData.AcknowledgeId,
                    ChildId = ""
                };

                bool isOta = commandData.CommandType.Equals("0x02", StringComparison.CurrentCultureIgnoreCase);
                int msgType = 5;
                if (isOta)
                {
                    //TODO : download and save ota file from url [commandData.Command]
                    ackDetails.Msg = "OTA updated successfully.";
                    ackDetails.Status = 7;
                    msgType = 11;
                }
                else
                {
                    ackDetails.Msg = "Device command received successfully.";
                    ackDetails.Status = 6;
                    msgType = 12;
                }

                var dataOta = JsonConvert.SerializeObject(ackDetails);

                sdkClient.SendAck(dataOta, msgType);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }

            return Task.CompletedTask;
        }
    }

    public class DeviceAckModel
    {
        [JsonProperty("ackId")]
        public string AckId { get; set; }
        [JsonProperty("st")]
        public int Status { get; set; }
        [JsonProperty("msg")]
        public string Msg { get; set; }
        [JsonProperty("childId")]
        public string ChildId { get; set; }
    }

    public class CommandData
    {
        [JsonProperty("cpId")]
        public string CpId { get; set; }

        [JsonProperty("guid")]
        public string Guid { get; set; } //Device Guid

        [JsonProperty("uniqueId")]
        public string UniqueId { get; set; } //Device unique id

        [JsonProperty("command")]
        public string Command { get; set; }

        [JsonProperty("value")]
        public string Value { get; set; }

        [JsonProperty("ack")]
        public bool Acknowledge { get; set; }

        [JsonProperty("ackId")]
        public string AcknowledgeId { get; set; }

        [JsonProperty("cmdType")]
        public string CommandType { get; set; }

        [JsonProperty("urls")]
        public object Urls { get; set; }
    }

    public class DeviceTwinCallBackModel
    {
        public TwinCollection Desired { get; set; }
        public TwinCollection Reported { get; set; }
        public string UniqueId { get; set; }
    }
}